#include <cstdarg>

#include "midi.hh"

using namespace std;

void MIDItrack::Add(unsigned char data)
{
    if(!(track.size() & (0x200-1)))
        track.reserve((track.size() + 0x200) & ~(0x200-1));
    track.push_back(data);
}
void MIDItrack::Flush()
{
    AddVarLen(delay); delay = 0;
}
void MIDItrack::AddVarLen(unsigned t)
{
    if(t>>21)Add(0x80 | ((t>>21) &127));
    if(t>>14)Add(0x80 | ((t>>14) &127));
    if(t>>7) Add(0x80 | ((t>>7)  &127));
    Add(t&127);
}
void MIDItrack::AddText(unsigned char n, const char* t)
{
    Flush();
    Add(0xFF);
    Add(n);
    Add(strlen(t));
    while(*t)Add(*t++);
}
void MIDItrack::AddData(unsigned char data)
{
    if(data & 0x80)
    {
    	Flush();
        if(data == running)return;
        running = data;
    }
    Add(data);
}
void MIDItrack::AddData(unsigned char nbytes, unsigned char data, ...)
{
    va_list ap;
    va_start(ap, data);
    for(;;)
    {
    	AddData(data);
    	if(nbytes <= 1)break;
    	--nbytes;
    	data = (unsigned char)(va_arg(ap, int));
    }
    va_end(ap);
}
void MIDItrack::AddMeta(unsigned char metatype, unsigned char nbytes, ...)
{
	Flush();
    Add(0xFF);
    Add(metatype);
    Add(nbytes);
    va_list ap;
    va_start(ap, nbytes);
    while(nbytes > 0)
    {
        unsigned char byte = (unsigned char)(va_arg(ap, int));
        Add(byte);
        --nbytes;
    }
    va_end(ap);
}


void MIDIfile::InitTrack(unsigned trackno)
{
    if(tracks.size() <= trackno)
    {
        tracks.reserve(16);
        tracks.resize(trackno+1);
    }
    
    tracks[trackno].Flush();
    
    // Meta 0x58 (misc settings):
        //      time signature: 4/2
        //      ticks/metro:    24
        //      32nd per 1/4:   8
    tracks[trackno].AddMeta(0x58, 4, 4,2, 24,8);
    
    // Meta 0x51 (tempo):
    tracks[trackno].AddMeta(0x51, 3, ((tempo>>16)&255),
                                     ((tempo>>8)&255),
                                     ((tempo)&255));
}
void MIDIfile::Done()
{
    AddLoopEnd();
    for(unsigned a=0; a<tracks.size(); ++a)
    {
        tracks[a].Flush();
        // Meta 0x2F (track end marker):
        tracks[a].AddMeta(0x2F, 0);
    }
    
    midi.clear();
    // Estimate the size of the MIDI file.
    unsigned midisize=10+2+2;
    for(unsigned a=0; a<tracks.size(); ++a)
        midisize += tracks[a].size() + 8;
    midi.reserve(midisize);
    
    for(unsigned c=0; c<10; ++c)
        midi.push_back(
            "MThd\0\0\0\6" // MIDI signature (MThd and number 6)
            "\0\1"         // Format 1
              [c]);
    
    // Track count
    midi.push_back(0);
    midi.push_back(tracks.size());
    
    midi.push_back((deltaticks>>8)&255);
    midi.push_back((deltaticks   )&255);
    
    for(unsigned a=0; a<tracks.size(); ++a)
    {
        for(unsigned c=0; c<4; ++c)
            midi.push_back("MTrk"[c]);
        
        unsigned tracklen = tracks[a].size();
        
        midi.push_back((tracklen>>24)&255);
        midi.push_back((tracklen>>16)&255);
        midi.push_back((tracklen>> 8)&255); 
        midi.push_back((tracklen    )&255);
        
        for(unsigned c=0; c<tracklen; ++c)
            midi.push_back(tracks[a][c]);
    }
}

void MIDIfile::KeyOn(unsigned trackno, unsigned ch, unsigned char note, unsigned char pressure)
{
    tracks[trackno].AddData(3, 0x90|ch, note, pressure);
}

void MIDIfile::KeyOff(unsigned trackno, unsigned ch, unsigned char note, unsigned char pressure)
{
    tracks[trackno].AddData(3, 0x80|ch, note, pressure);
}

void MIDIfile::KeyTouch(unsigned trackno, unsigned ch, unsigned char note, unsigned char pressure)
{
    tracks[trackno].AddData(3, 0xA0|ch, note, pressure);
}

void MIDIfile::Control(unsigned trackno, unsigned ch, unsigned char ctrl, unsigned char value)
{
    tracks[trackno].AddData(3, 0xB0|ch, ctrl, value);
}

void MIDIfile::Patch(unsigned trackno, unsigned ch, unsigned patchno)
{
    tracks[trackno].AddData(2, 0xC0|ch, patchno);
}

void MIDIfile::Wheel(unsigned trackno, unsigned ch, unsigned value)
{
    tracks[trackno].AddData(3, 0xE0|ch, value&127, (value>>7)&127);
}

void MIDIfile::Save(const char* filename)
{
    FILE* fp = fopen(filename, "wb");
    if(!fp) return;
    
    fwrite(getdata(), 1, size(), fp);
    fclose(fp);
}
